/* ----------------------------------------------------------------------------
 * Raima Database Manager
 *
 * Copyright (c) 2012 Raima Inc.,  All rights reserved.
 *
 * Use of this software, whether in source code format, or in executable,
 * binary object code form, is governed by the Raima LICENSE which
 * is fully described in the LICENSE.TXT file, included within this
 * distribution of files.
 * --------------------------------------------------------------------------*/
#include <stdio.h>

/* rdm.h is required in all files using the RDM Database */
#include "rdmsql.h"

/*! \file perf02.c
 * \brief Source code for the RDM Perf02 RDMSQL example
 */
 
/*!
 * \mainpage Perf02_RDMSQL
 * \author Daigoro F. Toyama, Raima Sr. Engineer
 * \date 2012
 *
 * \section Description
 * EXAMPLE - Perf02 RDMSQL
 *
 * The Perf02 example performs a number of atomic database operations
 * on a very simple database schema using the RDM SQL APIs.
 *
 * Each time you run this example
 *   * The database is initialized (removing all existing data)
 *   * 50,000 new records are created
 *   * Each record is read
 *   * Each record is updated
 *   * Each record is read again
 *   * Each record is deleted
 *   * Database is cleaned up
 *
 * \section Perf02_db Database
 * The DDL (Database Definition Language) specification for the database used
 * in this example is located in the file perf02.sdl. The example utilizes 
 * the SQL DDL syntax and specifies an on-disk database with a single 
 * table (simple).  The table has a single column (int_col).  There are no 
 * index definitions for this database.
 *
 * \section RDM SQL Functions Used
 * The following RDM functions are illustrated:
 *   * rsqlTFSInitEx/rsqlTFSTerm
 *   * rsqlAllocConn/rsqlFreeConn
 *   * rsqlOpenDB/rsqlCloseDB
 *   * rsqlInitDB
 *   * rsqlAllocStmt/rsqlFreeStmt
 *   * rsqlPrepare
 *   * rsqlBindParam
 *   * rsqlExecute
 *   * rsqlExecDirect
 *   * rsqlFetch
 *
 * They are fully documented in http://docs.raima.com/rdm/.
 */

/*! \brief Create a test database
 *
 */
static RSQL_ERRCODE create_database(void)
{
    RSQL_ERRCODE rsqlError;
    HCONN        hDbc;
    HSTMT        hStmt;
    int16_t      ii;

    const char *cmds[] = {
        "CREATE DATABASE perf02_db",
        "CREATE TABLE simple (int_col INTEGER)",
        "COMMIT"
    };

    (void) rsqlAllocConn(&hDbc);
    (void) rsqlAllocStmt(hDbc, &hStmt);

    /* Attempt to drop the database.  If the database does not already exist
     * this will fail and we simply ignore the failure for this sample. */
    (void) rsqlExecDirect(hStmt, "DROP DATABASE perf02_db");
    (void) rsqlExecDirect(hStmt, "COMMIT");

    for (ii = 0; ii < sizeof(cmds) / sizeof(cmds[0]); ii++)
    {
        rsqlError = rsqlExecDirect(hStmt, cmds[ii]);
        if (rsqlError != errSUCCESS)
            break;
    }

    if (rsqlError != errSUCCESS)
    {
        fprintf(stderr, "Error creating the perf02_db database\n");
        return rsqlError;
    }

    /* Close the newly created database */
    (void) rsqlCloseDB(hDbc, "perf02_db");

    (void) rsqlFreeStmt(hStmt);
    (void) rsqlFreeConn(hDbc);

    return errSUCCESS;
}

/*! \brief Display a message and the elapsed time to perform database operations
 *
 */
void display_elapsed_time(
    const char *msg,         /**< [in] Message to be display along with the elapsed time */
    uint64_t    start_time,  /**< [in] Start of elapsed time to display */
    uint64_t    end_time)    /**< [in] End of elapsed */
{
    uint64_t elapsed_time = end_time - start_time;
    printf("%s" UINT64_SPEC " millisecs\n", msg, elapsed_time/1000000);
}

/*! \brief Initialize the RDM SQL library for use in the perf02 example
 *
 * This function initializes the RDM Transactional File Server (TFS) to use 
 * the single-user non-transactional STANDALONE implementation.  It also creates
 * an RDM runtime task and opens the "perf02_db" database in exclusive ('x') mode.
 *
 * @return Returns a RSQL_ERRCODE value (errSUCCESS if successful)
 */
static RSQL_ERRCODE initialize(
    HCONN    *hDBC)   /**< [out] Pointer to the connection handle to be initialized */
{
    RSQL_ERRCODE rsqlError;
    uint64_t     start_time;
    uint64_t     end_time;

    *hDBC = NULL;

    /* Initialize the TFS to use the STANDALONE configuration (TFSS) */    
    start_time = psp_timeNanoSecs();

    rsqlError = create_database();
    if(rsqlError != errSUCCESS)
        return rsqlError;

    rsqlError = rsqlTFSInitEx(NULL, NULL, TFS_TYPE_STANDALONE);
    if(rsqlError != errSUCCESS)
    {
        fprintf(stderr, "Error initializing the TFS\n");
        return rsqlError;
    }

    /* Allocate the connection handle */
    rsqlError = rsqlAllocConn(hDBC);
    if(rsqlError != errSUCCESS)
    {
        fprintf(stderr, "Error allocating a connection handle\n");
        return rsqlError;
    }

    /* Open the perf02_db database in exclusive mode (does not require locks or tranactions) */
    rsqlError = rsqlOpenDB(*hDBC, "perf02_db", "x");
    if(rsqlError != errSUCCESS)
    {
        fprintf(stderr, "Error opening the perf02_db database\n");
        return rsqlError;
    }

    /* Initialize the database (creates new database files truncating any existing data) */    
    rsqlError = rsqlInitDB(*hDBC, "perf02_db");
    if (rsqlError != errSUCCESS)
    {
        printf("Error has occurred initializing database");
        return rsqlError;
    }

    end_time = psp_timeNanoSecs();  
    
    /* Display information */
    printf("Initializing with STANDALONE configuration\n");
    display_elapsed_time("Preparing a new database:\t\t\t", start_time, end_time);
    printf("\tOne, simple record type.\n");
    printf("\tOn disk\n");
    printf("\tOpened exclusively\n");

    return rsqlError;
}

/*! \brief Cleanup the RDM runtime library
 *
 * This functions closes all open databases and cleans up the RDM connection
 * handle used in the perf02 example.  It also terminates the STANDALONE TFS
 * used in the example. 
 *
 * @return Returns a RSQL_ERRCODE value (errSUCCESS if successful)
 */
static void cleanup(
    HCONN hDBC)   /**< [in] Connection handle to be closed/cleaned up */

{
    uint64_t     start_time;
    uint64_t     end_time;

    start_time = psp_timeNanoSecs();

    if(hDBC)
    {
        /* Close any open databases */
        rsqlCloseDB(hDBC, "perf02_db");

        /* Free the connection handle */
        rsqlFreeConn(hDBC);
    }

    /* Terminate the STANDALONE TFS */
    rsqlTFSTerm();

    end_time = psp_timeNanoSecs();

    display_elapsed_time("Cleanup:\t\t\t\t\t", start_time, end_time);
}

/*! \brief Insert 50,000 simple records into the perf02_db database
 *
 * This function adds 50,000 simple records into the on-disk perf database.  The
 * simple table contains only a single integer column with no index.  The function will
 * calculate and display the total time elapsed while inserting the records
 *
 * @return Returns a RSQL_ERRCODE value (errSUCCESS if successful)
 */
int32_t add_records(
    HCONN hDBC)         /**< [in] RDM SQL Connection handle */
{
    int32_t       rdm_status = S_OKAY;
    int32_t       ii;
    int32_t       int_col = 0;
    uint64_t      start_time;
    uint64_t      end_time;
    HSTMT         hSTMT;
    RSQL_ERRCODE  rsqlError;

    start_time = psp_timeNanoSecs();

    /* allocates insert statement */
    rsqlError = rsqlAllocStmt(hDBC, &hSTMT);
    if (rsqlError != errSUCCESS)
    {
        printf("Error has occured allocating insert statement");
        return rsqlError;
    }

    /* Prepare insert statement */
    rsqlError = rsqlPrepare(hSTMT,  "insert into simple values(?)");
    if (rsqlError != errSUCCESS)
    {
        rsqlFreeStmt(hSTMT);
        printf("Error has occurred preparing the insert statement");
        return rsqlError;
    }

    rsqlError = rsqlBindParam(hSTMT, 1, tINTEGER, &int_col, 0);
    if (rsqlError != errSUCCESS)
    {
        rsqlFreeStmt(hSTMT);
        printf("Error has occurred binding the insert statement parameter");
        return rsqlError;
    }

    /* Loop to add 50,000 new simple record */
    for(ii = 0; ii < 50000; ii++)
    {
        /* Set values for the new record */
        int_col = 1;

        /* Add the record to the database */
        rdm_status = rsqlExecute(hSTMT);
        if (rsqlError != errSUCCESS)
        {
            rsqlFreeStmt(hSTMT);
            printf("Error has occurred exeucting the insert statement");
            return rsqlError;
        }
    }
    rsqlFreeStmt(hSTMT);
    end_time = psp_timeNanoSecs();

    display_elapsed_time("Add 50,000 records:\t\t\t\t", start_time, end_time);
    return rdm_status;
}

/*! \brief Read all simple records and calculate the sum of the int_col column
 *
 * This function reads all simple records stored in the perf02_db database. The
 * function will calculate the sum of the int_col field from all simple rows
 * in the database.  Note that it actually would be more efficient to use the
 * build in SUM aggregate function than to manually fetch each row and 
 * calculate the sum in the function ("SELECT SUM(int_col) from simple"). The
 * function will also calculate and display the total time elapsed while 
 * reading the records.
 *
 * @return Returns a RSQL_ERRCODE value (errSUCCESS if successful)
 */
int32_t read_records(
    HCONN hDBC)         /**< [in] RDM SQL Connection Handle */
{
    uint64_t          total = 0;
    uint64_t          start_time;
    uint64_t          end_time;
    HSTMT             hSTMT;
    RSQL_ERRCODE      rsqlError;
    const RSQL_VALUE *result;
    uint16_t          numCols;

    start_time = psp_timeNanoSecs();

    /* allocates select statement */
    rsqlError = rsqlAllocStmt(hDBC, &hSTMT);
    if (rsqlError != errSUCCESS)
    {
        printf("Error has occured allocating select statement");
        return rsqlError;
    }

    /* Prepare select statement */
    rsqlError = rsqlExecDirect(hSTMT,  "select int_col from simple");
    if (rsqlError != errSUCCESS)
    {
        rsqlFreeStmt(hSTMT);
        printf("Error has occurred preparing the select statement");
        return rsqlError;
    }

    /* Loop through each simple record */
    while((rsqlError = rsqlFetch(hSTMT, &result, &numCols)) == errSUCCESS)
    {
        /* Add the value read to the running sum */
        total += result[0].vt.lv;
    }        

    /* We should break out of the loop when we get errNOMOREDATA */
    if (rsqlError == errNOMOREDATA)
        rsqlError = errSUCCESS;
    else
    {
        printf("Error has occurred fetching rows from the simple table\n");
        rsqlFreeStmt(hSTMT);
        return rsqlError;
    }

    rsqlFreeStmt(hSTMT);
    end_time = psp_timeNanoSecs();
    
    display_elapsed_time("Scanning, reading, summing 50,000 records:\t", start_time, end_time);
    printf("\tSum = " UINT64_SPEC "\n", total);
    return rsqlError;
}

/*! \brief Update all simple records in the database
 *
 * This function iterates updates all simple records stored in the perf02_db
 * database.  The function will calculate and display the total time elapsed
 * while deleting the records. 
 *
 * @return Returns a RSQL_ERRCODE value (errSUCCESS if successful)
 */
int32_t update_records(
    HCONN hDBC)         /**< [in] RDM SQL Connection Handle */
{
    uint64_t      start_time;
    uint64_t      end_time;
    HSTMT         hSTMT;
    RSQL_ERRCODE  rsqlError;

    start_time = psp_timeNanoSecs();

    /* allocates update statement */
    rsqlError = rsqlAllocStmt(hDBC, &hSTMT);
    if (rsqlError != errSUCCESS)
    {
        printf("Error has occured allocating update statement");
        return rsqlError;
    }

    /* Execute update statement */
    rsqlError = rsqlExecDirect(hSTMT,  "update simple set int_col = int_col + 1");
    if (rsqlError != errSUCCESS)
    {
        rsqlFreeStmt(hSTMT);
        printf("Error has occurred executing the update statement");
        return rsqlError;
    }
    rsqlFreeStmt(hSTMT);

    end_time = psp_timeNanoSecs();

    display_elapsed_time("Updating 50,000 records:\t\t\t", start_time, end_time);
    return rsqlError;
}

/*! \brief Delete all simple records from the database
 *
 * This function will delete all simple records stored in the perf02_db database.
 * The function will calculate and display the total time elapsed while 
 * deleting the records.
 *
 * @return Returns a RSQL_ERRCODE value (errSUCCESS if successful)
 */
int32_t delete_records(
    HCONN hDBC)         /**< [in]
 RDM task initialized with the perf02_db database open and as the current database */
{
    uint64_t      start_time;
    uint64_t      end_time;
    HSTMT         hSTMT;
    RSQL_ERRCODE  rsqlError;

    start_time = psp_timeNanoSecs();
    
    /* allocates delete statement */
    rsqlError = rsqlAllocStmt(hDBC, &hSTMT);
    if (rsqlError != errSUCCESS)
    {
        printf("Error has occured allocating delete statement");
        return rsqlError;
    }

    /* Execute delete statement */
    rsqlError = rsqlExecDirect(hSTMT,  "delete from simple");
    if (rsqlError != errSUCCESS)
    {
        rsqlFreeStmt(hSTMT);
        printf("Error has occurred executing the delete statement\n");
        return rsqlError;
    }

    rsqlFreeStmt(hSTMT);
    end_time = psp_timeNanoSecs();

    display_elapsed_time("Deleting 50,000 records:\t\t\t", start_time, end_time);
    return rsqlError;
}

/*! \brief Main function for perf02 example
 *
 * The function initializes the RDM SQL environment and runs the create, read,
 * update, and delete operations.
 *
 * @return Returns 0 on success, and 1 if there were any errors
 */
int32_t main(void)
{
    int32_t rdm_status;
    HCONN   hDBC;

    /* Display the example title */
    printf("%s", RDM_EXAMPLE_DESC("Perf02"));

    /* Initialize the TFS, task, and open/initialize the database */
    rdm_status = initialize(&hDBC);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Application is aborting due to initialization error\n");
        cleanup(hDBC);
        return 1;
    }

    /* Add 50,000 records to the database */
    rdm_status = add_records(hDBC);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Application is aborting due to error adding records\n");
        cleanup(hDBC);
        return 1;
    }

    /* Read all of the records in the database */
    rdm_status = read_records(hDBC);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Application is aborting due to error reading records\n");
        cleanup(hDBC);
        return 1;
    }

    /* Update all of the records in the database */
    rdm_status = update_records(hDBC);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Application is aborting due to error updating records\n");
        cleanup(hDBC);
        return 1;
    }

    /* Read all of the records in the database */
    rdm_status = read_records(hDBC);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Application is aborting due to error reading records\n");
        cleanup(hDBC);
        return 1;
    }

    /* Remove all of the records in the database */
    rdm_status = delete_records(hDBC);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Application is aborting due to error deleting records\n");
        cleanup(hDBC);
        return 1;
    }

    cleanup(hDBC);

    printf("\nPress Enter to terminate");
    getchar();

	return 0;
}
